/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.bitctrl.Constants;
import com.bitctrl.resource.Configuration;

import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.iav.fuzzylib.fuzzylib.FuzzyVariable;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.FuzzyWert;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.OdFuzzyWert;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdmsFahrBahnZustandFuzzy;
import de.bsvrz.sys.funclib.bitctrl.modell.AnmeldeException;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateEvent;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateListener;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.OnlineDatensatz;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsNiederschlagsIntensitaet;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UmfeldDatenMessStelle;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.onlinedaten.OdUfdmsGlaette;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.onlinedaten.OdUfdsNiederschlagsIntensitaet;

/**
 * Basisklasse für einen Testfall der Straßensubsegmentanalyse.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("nls")
public class TestfallRegelnFahrBahnZustand {

	/**
	 * Empänger für die Istwerte.
	 */
	private class Empfaenger implements DatensatzUpdateListener {

		@Override
		public void datensatzAktualisiert(final DatensatzUpdateEvent event) {
			if (!event.getDatum().isValid()) {
				return;
			}

			long zeitstempel;
			FuzzyVariable ist;

			zeitstempel = event.getDatum().getZeitstempel();
			ist = ((FuzzyWert) event.getDatum()).getFuzzyWert();
			if (ist.isValid()) {
				istWerte.put(zeitstempel, ist);
			}
			if (istWerte.size() >= sollWerte.size()) {
				synchronized (this) {
					notify();
				}
			}
		}

	}

	/** Der Name des Wertes der getestet wird. */
	private final String name;

	/** Der Logger der Klasse. */
	private final Logger log;

	/** Die Verbindung zur Datenbank. */
	private final Connection connection;

	/** Der Empfänger der Ist-Werte. */
	private final Empfaenger empfaenger;

	/** Der Datensatz des Fuzzywerts. */
	private final OnlineDatensatz<FuzzyWert> fuzzyWert;

	/** Ordnet die Sollwerte dem Zeitstempel an dem sie gesendet wurden zu. */
	private final SortedMap<Long, FuzzyVariable> sollWerte;

	/** Ordnet die Istwerte dem Zeitstempel an dem sie gesendet wurden zu. */
	private final SortedMap<Long, FuzzyVariable> istWerte;

	/** Der Datensatz des fuzzyfizierende Messwerts. */
	private final OdUfdsNiederschlagsIntensitaet messwertNi;

	/** Der Aspekt mit dem der Messwert gesendet wird. */
	private final Aspect messwertNiAsp;

	/** Enthält die Testdaten. */
	private final SortedMap<Long, OdUfdsNiederschlagsIntensitaet.Daten> testdatenNi;

	/** Der Datensatz des fuzzyfizierende Messwerts. */
	private final OdUfdmsGlaette messwertGlaette;

	/** Der Aspekt mit dem der Messwert gesendet wird. */
	private final Aspect messwertGlaetteAsp;

	/** Enthält die Testdaten. */
	private final SortedMap<Long, OdUfdmsGlaette.Daten> testdatenGlaette;

	/**
	 * Initialisiert das Objekt.
	 *
	 * @param ufdm
	 *            die Umfelddatenmessstelle, an der der Test durchgeführt wird.
	 * @param ufdsNi
	 *            der Umfelddatensensor für die Niederschlagsintensität, an dem
	 *            der Test durchgeführt wird. Der Sensor muss zur angegebenen
	 *            Messstelle gehören.
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	public TestfallRegelnFahrBahnZustand(final UmfeldDatenMessStelle ufdm,
			final UfdsNiederschlagsIntensitaet ufdsNi) throws SQLException {
		final Configuration konfig;
		final Level level;

		konfig = Configuration.getConfiguration();
		level = Level.parse(konfig.getString("debug.level"));
		log = Logger.getLogger(getClass().getName());
		log.setLevel(level);

		istWerte = new TreeMap<>();
		empfaenger = new Empfaenger();

		name = "FahrBahnZustand";
		fuzzyWert = ufdm.getOnlineDatensatz(OdUfdmsFahrBahnZustandFuzzy.class);

		messwertNi = ufdsNi
				.getOnlineDatensatz(OdUfdsNiederschlagsIntensitaet.class);
		messwertNiAsp = OdUfdsNiederschlagsIntensitaet.Aspekte.MessWertErsetzung
				.getAspekt();

		messwertGlaette = ufdm.getOnlineDatensatz(OdUfdmsGlaette.class);
		messwertGlaetteAsp = OdUfdmsGlaette.Aspekte.Prognose.getAspekt();

		connection = initConnection();
		testdatenNi = abfragenTestdatenNi();
		testdatenGlaette = abfragenTestdatenGlaette();
		sollWerte = abfragenSoll();
	}

	/**
	 * Initialisiert die Verbindung zur Datenbank.
	 *
	 * @return die offene Datenbankverbindung.
	 */
	private Connection initConnection() {
		Connection conn;

		log.fine(name + ": Lade Datenbanktreiber ...");
		try {
			Class.forName("org.hsqldb.jdbcDriver");
		} catch (final ClassNotFoundException ex) {
			throw new IllegalStateException(
					"Datenbanktreiber nicht gefunden: " + ex);
		}

		log.fine(name + ": Verbinde mit Datenbank ...");
		try {
			conn = DriverManager.getConnection(
					"jdbc:hsqldb:res:de/bsvrz/iav/subsegana/subsegana/prueffall7;shutdown=true",
					"sa", "");
		} catch (final SQLException ex) {
			throw new IllegalStateException(
					"Kann keine Verbindung zur Datenbank aufbauen: " + ex);
		}

		return conn;
	}

	/**
	 * Gibt die Liste der zu sendenden Testdaten zurück.
	 *
	 * @return die Testdaten.
	 */
	public SortedMap<Long, OdUfdsNiederschlagsIntensitaet.Daten> getTestdatenNi() {
		return testdatenNi;
	}

	/**
	 * Gibt die Liste der zu sendenden Testdaten zurück.
	 *
	 * @return die Testdaten.
	 */
	public SortedMap<Long, OdUfdmsGlaette.Daten> getTestdatenGlaette() {
		return testdatenGlaette;
	}

	/**
	 * Gibt die Liste der Sollwerte zurück.
	 *
	 * @return die Sollwerte.
	 */
	public SortedMap<Long, FuzzyVariable> getSollWerte() {
		return sollWerte;
	}

	/**
	 * Gibt die Liste der Istwerte zurück.
	 *
	 * @return die Istwerte.
	 */
	public SortedMap<Long, FuzzyVariable> getIstWerte() {
		return istWerte;
	}

	/**
	 * Führt den Test aus.
	 */
	public void ausfuehren() {
		int laenge;

		laenge = getTestdatenNi().size();
		if (laenge < getSollWerte().size()) {
			fail("Die Liste der Testdaten muss mindestens so lang sein wie die der Sollwerte.");
		}

		log.fine(name + ": Anmelden als Empfänger der Ist-Werte ...");
		fuzzyWert.addUpdateListener(OdFuzzyWert.Aspekte.Fuzzy.getAspekt(),
				empfaenger);

		// Testdaten senden.
		log.fine(name + ": Anmelden als Sender (Quelle) der Testdaten ...");
		messwertNi.setQuelle(messwertNiAsp, true);
		messwertGlaette.setQuelle(messwertGlaetteAsp, true);
		try {
			messwertNi.anmeldenSender(messwertNiAsp);
			messwertGlaette.anmeldenSender(messwertGlaetteAsp);
		} catch (final AnmeldeException ex) {
			ex.printStackTrace();
			fail("Konnte sich nicht als Sender für Testdaten anmelden: " + ex);
		}
		log.fine(name + ": Sende Testdaten ...");
		for (final long t : testdatenNi.keySet()) {
			final OdUfdsNiederschlagsIntensitaet.Daten datumNi;
			final OdUfdmsGlaette.Daten datumGlaette;

			datumNi = getTestdatenNi().get(t);
			datumNi.setZeitstempel(t);
			datumGlaette = getTestdatenGlaette().get(t);
			datumGlaette.setZeitstempel(t);
			log.finer("Sende Testdaten für den Zeitpunkt "
					+ com.bitctrl.util.Timestamp.absoluteTime(t) + " (" + t
					+ ") ...");
			try {
				messwertNi.sendeDaten(messwertNiAsp, datumNi);
				messwertGlaette.sendeDaten(messwertGlaetteAsp, datumGlaette);
			} catch (final DatensendeException ex) {
				ex.printStackTrace();
				fail("Konnte Testdaten nicht senden: " + ex);
			}
		}

		// Warten bis alle Antworten empfangen wurden.
		log.fine(name + ": Warte auf Eingang aller Ist-Werte ...");
		if (istWerte.size() < sollWerte.size()) {
			try {
				synchronized (empfaenger) {
					empfaenger.wait(Constants.MILLIS_PER_MINUTE);
				}
			} catch (final InterruptedException ex) {
				ex.printStackTrace();
			}
		}
	}

	/**
	 * Vergleicht die empfangen Ist-Werte mit den Soll-Werten.
	 */
	public void auswerten() {
		log.fine(name
				+ ": Vom Datenverteiler als Sender und Empfänger abmelden ...");
		fuzzyWert.removeUpdateListener(OdFuzzyWert.Aspekte.Fuzzy.getAspekt(),
				empfaenger);
		messwertNi.abmeldenSender(messwertNiAsp);

		log.fine(name + ": Vergleiche Soll- und Ist-Werte ...");
		for (final Long t : getSollWerte().keySet()) {
			final String zeitpunkt;
			final FuzzyVariable soll, ist;

			zeitpunkt = com.bitctrl.util.Timestamp.absoluteTime(t);
			log.finer("Vergleiche Soll- und Istwert zum Zeitpunkt " + zeitpunkt
					+ " (" + t + ") ...");

			soll = getSollWerte().get(t);
			ist = getIstWerte().get(t);
			assertEquals("Soll und Ist stimmen nicht überein: zeitpunkt="
					+ zeitpunkt, soll, ist);
		}
	}

	/**
	 * Gibt den Datensatz mit dem Messwert zurück.
	 *
	 * @return der Messwertdatensatz.
	 */
	public OdUfdsNiederschlagsIntensitaet getMesswertNi() {
		return messwertNi;
	}

	/**
	 * Gibt den Aspekt des Messwertdatensatzes zurück.
	 *
	 * @return der Aspekt.
	 */
	public Aspect getMesswertGlaetteAsp() {
		return messwertGlaetteAsp;
	}

	/**
	 * Gibt den Datensatz mit dem Messwert zurück.
	 *
	 * @return der Messwertdatensatz.
	 */
	public OdUfdmsGlaette getMesswertGlaette() {
		return messwertGlaette;
	}

	/**
	 * Gibt den Aspekt des Messwertdatensatzes zurück.
	 *
	 * @return der Aspekt.
	 */
	public Aspect getMesswertAspNi() {
		return messwertNiAsp;
	}

	/**
	 * Gibt die Verbindung zur Datenbank zurück.
	 *
	 * @return die Datenbankverbindung.
	 */
	public Connection getConnection() {
		return connection;
	}

	/**
	 * Gibt den Datensatz mit dem Fuzzy-Wert zurück.
	 *
	 * @return der Fuzzy-Wertdatensatz.
	 */
	public OnlineDatensatz<FuzzyWert> getFuzzyWert() {
		return fuzzyWert;
	}

	/**
	 * Gibt den Namen des Testfalls zurück.
	 *
	 * @return der Name des Testfalls.
	 */
	public String getName() {
		return name;
	}

	/**
	 * Gibt den Logger der Klasse zurück.
	 *
	 * @return der Logger.
	 */
	public Logger getLog() {
		return log;
	}

	/**
	 * Ruft die Sollwerte aus der Datenbank ab.
	 *
	 * @return die Sollwerte.
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	protected SortedMap<Long, FuzzyVariable> abfragenSoll()
			throws SQLException {
		final Statement stat;
		final ResultSet rs;
		final String sql;
		final SortedMap<Long, FuzzyVariable> daten;

		getLog().fine(getName() + ": Lese Sollwerte aus Datenbank ...");

		daten = new TreeMap<>();
		stat = getConnection().createStatement();
		sql = "SELECT * FROM regeln WHERE testname LIKE '" + getName() + "'";
		rs = stat.executeQuery(sql);

		while (rs.next()) {
			final FuzzyVariable fv;
			final String sollText;
			final String[] sollwerte;
			final List<String> termnamen;
			final Map<String, Float> terme;
			final Timestamp zeitstempel;

			if (!rs.getString("wertname").equals(getName())) {
				continue;
			}

			zeitstempel = rs.getTimestamp("zeitstempel");
			sollText = rs.getString("soll");
			if (sollText == null) {
				continue;
			}
			sollwerte = sollText.split("\\|");
			termnamen = new ArrayList<>();
			terme = new HashMap<>();
			for (final String t : sollwerte) {
				String termname;

				termname = t.split(":")[0];
				termnamen.add(termname);
				terme.put(termname, Float.valueOf(t.split(":")[1]));
			}

			fv = new FuzzyVariable(getName(),
					termnamen.toArray(new String[termnamen.size()]));
			for (final String t : termnamen) {
				fv.setTerm(t, terme.get(t));
			}
			daten.put(zeitstempel.getTime(), fv);
		}

		return daten;
	}

	/**
	 * Ruft die Testdaten aus der Datenbank ab.
	 *
	 * @return die Testdaten.
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	protected SortedMap<Long, OdUfdsNiederschlagsIntensitaet.Daten> abfragenTestdatenNi()
			throws SQLException {
		final Statement stat;
		final ResultSet rs;
		final String sql;
		final SortedMap<Long, OdUfdsNiederschlagsIntensitaet.Daten> daten;

		getLog().fine(getName() + ": Lese Testdaten aus Datenbank ...");

		daten = new TreeMap<>();
		stat = getConnection().createStatement();
		sql = "SELECT * FROM regeln WHERE testname LIKE '" + getName() + "'";
		rs = stat.executeQuery(sql);

		while (rs.next()) {
			final OdUfdsNiederschlagsIntensitaet.Daten datum;
			final long zeitstempel;
			final Double wert;
			final String wertname;

			wertname = rs.getString("wertname");
			if (!wertname.equals("NiederschlagsIntensität")) {
				continue;
			}

			// Falls bereits vorhanden, Datensatz bestimmen, sonst anlegen
			zeitstempel = rs.getTimestamp("zeitstempel").getTime();
			if (daten.containsKey(zeitstempel)) {
				datum = daten.get(zeitstempel);
			} else {
				datum = getMesswertNi().erzeugeDatum();
				daten.put(zeitstempel, datum);
			}

			// Wert im Datensatz setzen, wenn ungleich null
			wert = rs.getDouble("messwert");
			if (!rs.wasNull()) {
				datum.setWert(wertname, wert);
			}
		}

		return daten;
	}

	/**
	 * Ruft die Testdaten aus der Datenbank ab.
	 *
	 * @return die Testdaten.
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	protected SortedMap<Long, OdUfdmsGlaette.Daten> abfragenTestdatenGlaette()
			throws SQLException {
		final Statement stat;
		final ResultSet rs;
		final String sql;
		final SortedMap<Long, OdUfdmsGlaette.Daten> daten;

		getLog().fine(getName() + ": Lese Testdaten aus Datenbank ...");

		daten = new TreeMap<>();
		stat = getConnection().createStatement();
		sql = "SELECT * FROM regeln WHERE testname LIKE '" + getName() + "'";
		rs = stat.executeQuery(sql);

		while (rs.next()) {
			final OdUfdmsGlaette.Daten datum;
			final long zeitstempel;
			final Double wert;
			final String wertname;

			wertname = rs.getString("wertname");
			if (!wertname.equals("Glätte")) {
				continue;
			}

			// Falls bereits vorhanden, Datensatz bestimmen, sonst anlegen
			zeitstempel = rs.getTimestamp("zeitstempel").getTime();
			if (daten.containsKey(zeitstempel)) {
				datum = daten.get(zeitstempel);
			} else {
				datum = getMesswertGlaette().erzeugeDatum();
				daten.put(zeitstempel, datum);
			}

			// Wert im Datensatz setzen, wenn ungleich null
			wert = rs.getDouble("messwert");
			if (!rs.wasNull()) {
				datum.setWert(wertname, wert);
			}
		}

		return daten;
	}

}
